// import type { worldline_id_t } from '../worldlines/model'
// import { annotation_id_t } from '../annotations/model'

// export type AnnotationWindowState_t = {
//   x: number,
//   y: number,
//   z: number,
//   c_idx: number,
//   t_idx: number,
//   shape_x: number,
//   shape_y: number,
//   shape_z: number,
//   views: string[],
//   shape_t: number,
//   scale_x: number,
//   scale_y: number,
//   scale_z: number,
//   selected_annotation: annotation_id_t | null,
//   visible_worldlines: worldline_id_t[],
//   selected_worldline: worldline_id_t | null,
//   fill_circles: boolean,
//   rpc_list: string[],
//   rpc_keybindings: {[key: string]: number}
//   rpc_args: {[key: string]: string}
// }