__version__ = '1.0.3'
# from pathlib import Path
# import subprocess
#
# __version__ = subprocess.check_output(
#     ["git", "describe", "--tags"],
#     cwd=Path(__file__).parent.parent.absolute()
# ).strip().decode('UTF-8')
